#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.exploration_techniques"  # pylint:disable=redefined-builtin

import logging
import os
import unittest

import angr

from ..common import broken, bin_location, do_trace, load_cgc_pov, slow_test, skip_if_not_linux


def tracer_cgc(
    filename,
    test_name,
    stdin,
    copy_states=False,
    follow_unsat=False,
    read_strategies=None,
    write_strategies=None,
    add_options=None,
    remove_options=None,
    syscall_data=None,
    symbolic_stdin=True,
):
    p = angr.Project(filename)
    p.simos.syscall_library.update(angr.SIM_LIBRARIES["cgcabi_tracer"])

    trace, magic, crash_mode, crash_addr = do_trace(p, test_name, stdin)
    s = p.factory.entry_state(
        mode="tracing",
        stdin=angr.SimFileStream,
        flag_page=magic,
        add_options=add_options,
        remove_options=remove_options,
    )
    if read_strategies is not None:
        s.memory.read_strategies = read_strategies
    if write_strategies is not None:
        s.memory.write_strategies = write_strategies
    s.preconstrainer.preconstrain_file(stdin, s.posix.stdin, True)

    simgr = p.factory.simulation_manager(s, hierarchy=None, save_unconstrained=crash_mode)
    t = angr.exploration_techniques.Tracer(
        trace,
        crash_addr=crash_addr,
        keep_predecessors=1,
        copy_states=copy_states,
        follow_unsat=follow_unsat,
        syscall_data=syscall_data,
    )
    if add_options is not None and angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL in add_options:
        if symbolic_stdin:
            fd_data = {0: (stdin, b"\x01" * len(stdin))}
        else:
            fd_data = {0: (stdin, b"\x00" * len(stdin))}

        t.set_fd_data(fd_data)

    simgr.use_technique(t)
    simgr.use_technique(angr.exploration_techniques.Oppologist())

    return simgr, t


def trace_cgc_with_pov_file(
    binary: str,
    test_name: str,
    pov_file: str,
    output_initial_bytes: bytes,
    copy_states=False,
    read_strategies=None,
    write_strategies=None,
    add_options=None,
    remove_options=None,
    syscall_data=None,
    symbolic_stdin=True,
):
    assert os.path.isfile(pov_file)
    pov = load_cgc_pov(pov_file)
    trace_result = tracer_cgc(
        binary,
        test_name,
        b"".join(pov.writes),
        copy_states,
        read_strategies=read_strategies,
        write_strategies=write_strategies,
        add_options=add_options,
        remove_options=remove_options,
        syscall_data=syscall_data,
        symbolic_stdin=symbolic_stdin,
    )
    simgr = trace_result[0]
    simgr.run()
    assert "traced" in simgr.stashes
    assert len(simgr.traced) == 1
    stdout_dump = simgr.traced[0].posix.dumps(1)
    assert stdout_dump.startswith(output_initial_bytes)


def tracer_linux(filename, test_name, stdin, add_options=None, remove_options=None):
    p = angr.Project(filename)

    trace, _, crash_mode, crash_addr = do_trace(
        p,
        test_name,
        stdin,
        ld_linux=p.loader.linux_loader_object.binary,
        library_path={os.path.dirname(obj.binary) for obj in p.loader.all_elf_objects},
        record_stdout=True,
    )
    s = p.factory.full_init_state(
        mode="tracing",
        stdin=angr.SimFileStream,
        add_options=add_options,
        remove_options=remove_options,
    )
    s.preconstrainer.preconstrain_file(stdin, s.posix.stdin, True)

    simgr = p.factory.simulation_manager(s, hierarchy=None, save_unconstrained=crash_mode)
    t = angr.exploration_techniques.Tracer(trace, crash_addr=crash_addr)
    simgr.use_technique(t)
    simgr.use_technique(angr.exploration_techniques.Oppologist())

    return simgr, t


class TestTracer(unittest.TestCase):
    def test_recursion(self):
        blob = bytes.fromhex(
            "00aadd114000000000000000200000001d0000000005000000aadd2a1100001d0000000001e8030000aadd21118611b3b3b3b3b3e3b1b"
            "1b1adb1b1b1b1b1b1118611981d8611"
        )
        fname = os.path.join(bin_location, "tests", "cgc", "NRFIN_00075")

        simgr, _ = tracer_cgc(fname, "tracer_recursion", blob)
        simgr.run()

        assert simgr.crashed
        assert simgr.crashed[0].solver.symbolic(simgr.crashed[0].regs.ip)

    @broken
    @slow_test
    def test_cache_stall(self):
        # test a valid palindrome
        b = os.path.join(bin_location, "tests", "cgc", "CROMU_00071")
        blob = bytes.fromhex(
            "0c0c492a53acacacacacacacacacacacacac000100800a0b690e0aef6503697d660a0059e20afc0a0a332f7d66660a0059e20afc0a0a3"
            "32f7fffffff16fb1616162516161616161616166a7dffffff7b0e0a0a6603697d660a0059e21c"
        )

        simgr, tracer = tracer_cgc(b, "tracer_cache_stall", blob)
        simgr.run()

        crash_path = tracer.predecessors[-1]
        crash_state = simgr.crashed[0]

        assert crash_path is not None
        assert crash_state is not None

        # load it again
        simgr, tracer = tracer_cgc(b, "tracer_cache_stall", blob)
        simgr.run()

        crash_path = tracer.predecessors[-1]
        crash_state = simgr.one_crashed

        assert crash_path is not None
        assert crash_state is not None

    @skip_if_not_linux
    def test_manual_recursion(self):
        b = os.path.join(bin_location, "tests", "cgc", "CROMU_00071")
        with open(os.path.join(bin_location, "tests_data", "crash2731"), "rb") as fh:
            blob = fh.read()

        simgr, tracer = tracer_cgc(b, "tracer_manual_recursion", blob)
        simgr.run()

        crash_path = tracer.predecessors[-1]
        crash_state = simgr.one_crashed

        assert crash_path is not None
        assert crash_state is not None

    def test_cgc_receive_unicorn_native_interface(self):
        """
        Test if unicorn native interface handles CGC receive syscall correctly. Receives with symbolic arguments also
        tested.
        """

        binary = os.path.join(bin_location, "tests", "cgc", "KPRCA_00038")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "KPRCA_00038_POV_00000.xml")
        output_initial_bytes = b""
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
            angr.options.UNICORN_HANDLE_SYMBOLIC_SYSCALLS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_cgc_receive_unicorn_native_interface",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    def test_cgc_receive_unicorn_native_interface_rx_bytes(self):
        """
        Test rx_bytes is correctly handled by unicorn native interface's CGC receive: update only if non-null
        """

        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00012")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00012_POV_00000.xml")
        output_initial_bytes = b""
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_cgc_receive_unicorn_native_interface_rx_bytes",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    def test_cgc_random_syscall_handling_native_interface(self):
        """
        Test if random syscall is correctly handled in native interface. Random with symbolic arguments also tested.
        """

        binary = os.path.join(bin_location, "tests", "cgc", "KPRCA_00011")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "KPRCA_00011_POV_00000.xml")
        output_file = os.path.join(bin_location, "tests_data", "cgc_povs", "KPRCA_00011_stdout.txt")
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_CGC_RANDOM_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
            angr.options.UNICORN_HANDLE_SYMBOLIC_SYSCALLS,
        }

        rand_syscall_data = {
            "random": [
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
                (65, 1),
                (16705, 2),
                (16705, 2),
            ]
        }
        with open(output_file, "rb") as fh:
            output_bytes = fh.read()

        trace_cgc_with_pov_file(
            binary,
            "tracer_cgc_receive_unicorn_native_interface_rx_bytes",
            pov_file,
            output_bytes,
            add_options=add_options,
            syscall_data=rand_syscall_data,
        )

    def test_cgc_se1_palindrome_raw(self):
        b = os.path.join(bin_location, "tests", "cgc", "sc1_0b32aa01_01")
        # test a valid palindrome

        simgr, _ = tracer_cgc(b, "tracer_cgc_se1_palindrome_raw_nocrash", b"racecar\n")
        simgr.run()

        # make sure the heap base is correct and hasn't been altered from the default
        assert "traced" in simgr.stashes
        assert simgr.traced[0].cgc.allocation_base == 0xB8000000

        # make sure there is no crash state
        assert not simgr.crashed

        # make sure angr modeled the correct output
        stdout_dump = simgr.traced[0].posix.dumps(1)
        assert stdout_dump.startswith(
            b"\nWelcome to Palindrome Finder\n\n"
            b"\tPlease enter a possible palindrome: "
            b"\t\tYes, that's a palindrome!\n\n"
            b"\tPlease enter a possible palindrome: "
        )
        # make sure there were no 'Nope's from non-palindromes
        assert b"Nope" not in stdout_dump

        # now test crashing input
        simgr, _ = tracer_cgc(b, "tracer_cgc_se1_palindrome_raw_yescrash", b"A" * 129)
        simgr.run()

        assert simgr.crashed

    def test_concrete_execution_in_native_interface(self):
        """
        Test if concrete execution without any symbolic bytes is done correctly when receive syscall is handled in
        native interface
        """

        binary = os.path.join(bin_location, "tests", "cgc", "KPRCA_00052")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "KPRCA_00052_POV_00000.xml")
        output_initial_bytes = (
            b"Enter system password: \nWelcome to the CGC Pizzeria order management system.\n1. Input Order\n"
            b"2. Update Order\n3. View One Orders\n4. View All Orders\n5. Delete Order\n6. Clear All Orders\n"
            b"7. Logout\n"
            b"Choice: Enter Pickup Name: Choose what the kind of pizza\n1. Pizza Pie - The classic!\n"
            b"2. Pizza Sub - All the fun, on a bun\n3. Pizza Bowl - Our own twist\nChoice: Select Size\n1. Small\n"
            b"2. Medium\n3. Large\nChoice: Successfully added a new Pizza Pie!\nSelect an option:\n1. Add Toppings\n"
            b"2. Remove Toppings\n3. Add Sauce\n4. Remove Sauce\n5. Finished With Pizza\nChoice: Successfully added "
            b"pizza!"
            b"\n1. Add another Pizza\n2. Quit\nChoice: 0. Cancel\n==================================================\n"
            b"  "
            b"Item #1. Classic Pizza Pie, Size: SMALL\n    Selected Toppings\n\tNone\n    Sauce on the side\n\tNone\n"
            b"--------------------------------------\n\t\tCalories: 1000\n\t\tCarbs   : 222\n\nPizza length... = 1\n"
            b"\t\t"
            b"Estimated wait time: 36 minute(s)\n==================================================\nChoice: "
            b"Removed Item #1\n1. Add another Pizza\n2. Quit\nChoice: Order successfully added!\n1. Input Order\n"
            b"2. Update Order\n3. View One Orders\n4. View All Orders\n5. Delete Order\n6. Clear All Orders\n7. "
            b"Logout\n"
            b"Choice: 1 - pov: Ordered 0 pizza(s)\n==================================================\n"
            b"--------------------------------------\n\t\tCalories: 0\n\t\tCarbs   : 0\n\n"
        )
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary,
            "concrete_execution_in_native_interface",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
            symbolic_stdin=False,
        )

    def test_d_flag_and_write_write_conflict_in_unicorn(self):
        """
        Check if d flag is handled correctly in unicorn native interface and write-write conflicts do not occur when
        re-executing symbolic instructions
        """

        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00008")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00008_POV_00000.xml")
        output_initial_bytes = (
            b"> You logged in.\n> First name: Last name: User name: Birthdate (mm/dd/yy hh:mm:ss): "
            b"Date is: 12/21/1983 5:43:21\nData added, record 0\n"
            b"> Enter search express (firstname or fn, lastname or ln, username or un, birthdate or bd,"
            b" operators ==, !=, >, <, AND and OR):\n"
        )
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_d_flag_and_write_write_conflict_in_unicorn",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    def test_empty_reexecute_block_remove_in_unicorn_native_interface(self):
        """
        Test if blocks with no symbolic instructions are removed from re-execution list in unicorn native interface.
        Re-execute instruction list of a block can become empty when all of them are removed when performing memory
        writes. See handle_write in unicorn native interface.
        """

        binary = os.path.join(bin_location, "tests", "cgc", "KPRCA_00052")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "KPRCA_00052_POV_00000.xml")
        output_initial_bytes = (
            b"Enter system password: \nWelcome to the CGC Pizzeria order management system.\n1. Input Order\n"
            b"2. Update Order\n3. View One Orders\n4. View All Orders\n5. Delete Order\n6. Clear All Orders\n"
            b"7. Logout\n"
            b"Choice: Enter Pickup Name: Choose what the kind of pizza\n1. Pizza Pie - The classic!\n"
            b"2. Pizza Sub - All the fun, on a bun\n3. Pizza Bowl - Our own twist\nChoice: Select Size\n1. Small\n"
            b"2. Medium\n3. Large\nChoice: Successfully added a new Pizza Pie!\nSelect an option:\n1. Add Toppings\n"
            b"2. Remove Toppings\n3. Add Sauce\n4. Remove Sauce\n5. Finished With Pizza\nChoice: Successfully added "
            b"pizza!"
            b"\n1. Add another Pizza\n2. Quit\nChoice: 0. Cancel\n=================================================="
            b"\n  "
            b"Item #1. Classic Pizza Pie, Size: SMALL\n    Selected Toppings\n\tNone\n    Sauce on the side\n\tNone\n"
            b"--------------------------------------\n\t\tCalories: 1000\n\t\tCarbs   : 222\n\nPizza length... = 1"
            b"\n\t\t"
            b"Estimated wait time: 36 minute(s)\n==================================================\nChoice: "
            b"Removed Item #1\n1. Add another Pizza\n2. Quit\nChoice: Order successfully added!\n1. Input Order\n"
            b"2. Update Order\n3. View One Orders\n4. View All Orders\n5. Delete Order\n6. Clear All Orders\n7. "
            b"Logout\n"
            b"Choice: 1 - pov: Ordered 0 pizza(s)\n==================================================\n"
            b"--------------------------------------\n\t\tCalories: 0\n\t\tCarbs   : 0\n\n"
        )
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_empty_reexecute_block_remove_in_unicorn_native_interface",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    def test_symbolic_sized_receives(self):
        b = os.path.join(bin_location, "tests", "cgc", "CROMU_00070")

        simgr, _ = tracer_cgc(b, "tracer_symbolic_sized_receives", b"hello")
        simgr.run()

        assert not simgr.crashed
        assert "traced" in simgr.stashes

        simgr, _ = tracer_cgc(b, "tracer_symbolic_sized_receives_nulls", b"\0" * 20)
        simgr.run()

        assert not simgr.crashed
        assert "traced" in simgr.stashes

    def test_allocation_base_continuity(self):
        correct_out = (
            b"prepare for a challenge\nb7fff000\nb7ffe000\nb7ffd000\nb7ffc000\nb7ffb000\nb7ffa000\nb7ff9000\nb7ff8000\n"
            b"b7ff7000\nb7ff6000\nb7ff5000\nb7ff4000\nb7ff3000\nb7ff2000\nb7ff1000\nb7ff0000\nb7fef000\nb7fee000\n"
            b"b7fed000\nb7fec000\ndeallocating b7ffa000\na: b7ffb000\nb: b7fff000\nc: b7ff5000\nd: b7feb000\n"
            b"e: b7fe8000\n"
            b"e: b7fa8000\na: b7ffe000\nb: b7ffd000\nc: b7ff7000\nd: b7ff6000\ne: b7ff3000\ne: b7f68000\nallocate: 3\n"
            b"a: b7fef000\n"
        )

        b = os.path.join(bin_location, "tests", "i386", "cgc_allocations")

        simgr, _ = tracer_cgc(b, "tracer_allocation_base_continuity", b"")
        simgr.run()

        assert simgr.traced[0].posix.dumps(1) == correct_out

    def test_crash_addr_detection(self):
        b = os.path.join(bin_location, "tests", "i386", "call_symbolic")

        simgr, _ = tracer_cgc(b, "tracer_crash_addr_detection", b"A" * 700)
        simgr.run()

        assert simgr.crashed
        assert simgr.crashed[0].solver.symbolic(simgr.crashed[0].regs.ip)

    @skip_if_not_linux
    def test_fauxware(self):
        b = os.path.join(bin_location, "tests", "x86_64", "fauxware")
        simgr, _ = tracer_linux(b, "tracer_fauxware", b"A" * 18, remove_options={angr.options.CPUID_SYMBOLIC})
        simgr.run()

        assert "traced" in simgr.stashes

    def test_rollback_on_symbolic_conditional_exit(self):
        """
        Test if state is correctly rolled back to before start of block in case block cannot be executed in unicorn
        engine because exit condition is symbolic
        """
        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00043")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00043_POV_00000.xml")
        output_initial_bytes = [
            b"Network type: Broadcast",
            b"Source Address: 0x962B175B",
            b"Network type: Endpoint",
            b"Source Address: 0x321B00B0",
            b"Destination Address: 0xACF70019",
            b"Final Statistics:",
            b"\tTotal Packets: 6",
            b"\tStart Time: 0x5552C470",
            b"\tEnd Time: 0x54CAF0B0",
            b"\tLargest Packet: 0",
            b"\tSmallest Packet: 0",
            b"\tNumber of malformed packets: 0",
            b"\tNumber of packets shown 6",
            b"Option Headers:",
            b"This content has not been modified from the original",
            b"Capturing Authority: Network Provider",
            b"Capture Date: bKQcAXJJEqCSPmrIlRy",
            b"Capturing Authority: Employer\n",
        ]
        trace_cgc_with_pov_file(
            binary,
            "tracer_rollback_on_symbolic_conditional_exit",
            pov_file,
            b"\n".join(output_initial_bytes),
        )

    def test_floating_point_memory_reads(self):
        """
        Test float point memory reads in which bytes longer than architecture width are read in a single memory read
        hook in unicorn. The other related case is when such reads are split across multiple reads. This is tested in
        b01lersctf2020 little engine solver
        """
        binary = os.path.join(bin_location, "tests", "cgc", "NRFIN_00027")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "NRFIN_00027_POV_00000.xml")
        output = b"\x00" * 36
        trace_cgc_with_pov_file(
            binary,
            "tracer_floating_point_memory_reads",
            pov_file,
            output,
            read_strategies=[angr.concretization_strategies.SimConcretizationStrategyAny(exact=True)],
            write_strategies=[angr.concretization_strategies.SimConcretizationStrategyAny(exact=True)],
        )

    def test_fdwait_fds(self):
        """
        Test fdwait working with appropriate bit order for read/write fds
        """
        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00029")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00029_POV_00000.xml")
        output = [
            b"For what material would you like to run this simulation?",
            b"  1. Air",
            b"  2. Aluminum",
            b"  3. Copper",
            b"  4. Custom\nSelection: ",
        ]
        trace_cgc_with_pov_file(binary, "tracer_floating_point_memory_reads", pov_file, b"\n".join(output))

    def test_non_zero_offset_subregister_dependency_saving_unicorn_native_interface(self):
        """
        Test if concrete register dependencies of symbolic instructions are saved correctly in unicorn native interface
        for re-executing
        """

        binary = os.path.join(bin_location, "tests", "cgc", "KPRCA_00028")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "KPRCA_00028_POV_00000.xml")
        output_initial_bytes = b"Welcome to the SLUR REPL. Type an expression to evaluate it.\n> "
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_non_zero_offset_subregister_dependency_saving_unicorn_native_interface",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    def test_saving_dependencies_of_last_instruction_of_block_in_unicorn_native_interface(self):
        """
        Test if dependencies of last instruction in a basic block are saved in unicorn native interface
        """

        binary = os.path.join(bin_location, "tests", "cgc", "NRFIN_00026")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "NRFIN_00026_POV_00000.xml")
        output_initial_bytes = (
            b"Starting dissection...\n\n\n====New Packet====\n\n\n===rofl===\n\n\n===rachiometersuprachoroid==="
            b"\n301478991"
            b"\nString display will be handled in v4.\n1\nString display will be handled in v4.\n0\n1\n"
            b"LV type will be handled in v4.\n3582705152\nString display will be handled in v4.\n"
            b"LV type will be handled in v4.\n190\n0\n===trolololo===\n"
        )
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_saving_dependencies_of_last_instruction_of_block_in_unicorn_native_interface",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    @slow_test
    def test_sseround_register_dependency_unicorn_native_interface(self):
        """
        Test if value of SSEROUND VEX register is saved correctly when it is a dependency of an instruction that needs
        to be re-executed. Takes about 10 minutes.
        """

        binary = os.path.join(bin_location, "tests", "cgc", "NRFIN_00021")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "NRFIN_00021_POV_00000.xml")
        output_initial_bytes = b""
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_sseround_register_dependency_unicorn_native_interface",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    def test_concretize_unsupported_vex_irops(self):
        """
        Test tracing with concretizing unsupported VEX IR Ops
        """

        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00020")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00020_POV_00000.xml")
        output = (
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x15\x00\x00\x00\x00\x00\x00\x00\x15"
            + b"\x00\x00\x00\x00\x00\x00\x00\x15\x00\x00\x00\x00\x00\x00\x00"
        )
        add_options = {
            angr.options.UNSUPPORTED_FORCE_CONCRETIZE,
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
        }
        trace_cgc_with_pov_file(
            binary, "tracer_concretize_unsupported_vex_ops", pov_file, output, add_options=add_options
        )

    def test_skip_some_symbolic_memory_writes(self):
        """
        Test symbolic memory write skipping in SimEngineUnicorn during tracing
        This test doesn't actually check if instruction was skipped. It checks if tracing is successful
        """

        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00023")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00023_POV_00000.xml")
        output_initial_bytes = [
            b"",
            b"C - Change Diver Info",
            b"L - Log a New Dive",
            b"D - Download Dive Data",
            b"E - Edit Dives",
            b"P - Print Dive Logs",
            b"R - Remove Dives",
            b"S - Diver Statistics",
            b"X - Exit Application",
            b":",
            b"",
            b"Dive Log is empty",
            b"",
            b"C - Change Diver Info",
            b"L - Log a New Dive",
            b"D - Download Dive Data",
            b"E - Edit Dives",
            b"P - Print Dive Logs",
            b"R - Remove Dives",
            b"S - Diver Statistics",
            b"X - Exit Application",
            b":",
            b"",
            b"Dive Log is empty",
            b"",
            b"C - Change Diver Info",
            b"L - Log a New Dive",
            b"D - Download Dive Data",
            b"E - Edit Dives",
            b"P - Print Dive Logs",
            b"R - Remove Dives",
            b"S - Diver Statistics",
            b"X - Exit Application",
            b":",
            (
                b"Dive Site: Date: Time: Location (area/city): Max Depth in ft: Avg Depth in ft: "
                b"Dive Duration (mins): O2 Percentage: Pressure In (psi): Pressure Out (psi): "
            ),
            b"C - Change Diver Info",
            b"L - Log a New Dive",
            b"D - Download Dive Data",
            b"E - Edit Dives",
            b"P - Print Dive Logs",
            b"R - Remove Dives",
            b"S - Diver Statistics",
            b"X - Exit Application",
            b":",
            (
                b"Dive Site: Date: Time: Location (area/city): Max Depth in ft: Avg Depth in ft: "
                b"Dive Duration (mins): O2 Percentage: Pressure In (psi): Pressure Out (psi): "
            ),
            b"C - Change Diver Info",
            b"L - Log a New Dive",
            b"D - Download Dive Data",
            b"E - Edit Dives",
            b"P - Print Dive Logs",
            b"R - Remove Dives",
            b"S - Diver Statistics",
            b"X - Exit Application",
            b":",
            (
                b"First Name: Last Name: Street: City: State: Zip Code: Phone Number: PADI Diver Number: "
                b"PADI Cert Date: "
            ),
            b"     Name: ",
        ]
        trace_cgc_with_pov_file(
            binary,
            "tracer_skip_some_symbolic_memory_writes",
            pov_file,
            b"\n".join(output_initial_bytes),
        )

    def test_subregister_tainting(self):
        """
        Tests for subregister tainting: taint only bytes of subregister and not entire register
        """
        binary = os.path.join(bin_location, "tests", "cgc", "KPRCA_00028")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "KPRCA_00028_POV_00000.xml")
        output_initial_bytes = b"Welcome to the SLUR REPL. Type an expression to evaluate it.\n> "
        trace_cgc_with_pov_file(binary, "tracer_subregister_tainting", pov_file, output_initial_bytes)

    def test_symbolic_memory_dependencies_liveness(self):
        """
        Tests for liveness of symbolic memory dependencies when re-executing symbolic instructions in SimEngineUnicorn
        NRFIN_00036
        """
        binary = os.path.join(bin_location, "tests", "cgc", "NRFIN_00036")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "NRFIN_00036_POV_00000.xml")
        output_initial_bytes = b"New budget created!\nNew budget created!\nNew budget created!\nNew budget created!\n"
        trace_cgc_with_pov_file(
            binary,
            "tracer_symbolic_memory_dependencies_liveness",
            pov_file,
            output_initial_bytes,
        )

        # CROMU_00008
        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00008")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00008_POV_00000.xml")
        output_initial_bytes = (
            b"> You logged in.\n> First name: Last name: User name: Birthdate (mm/dd/yy hh:mm:ss): "
            b"Date is: 12/21/1983 5:43:21\nData added, record 0\n"
            b"> Enter search express (firstname or fn, lastname or ln, username or un, birthdate or bd,"
            b" operators ==, !=, >, <, AND and OR):\n"
        )
        trace_cgc_with_pov_file(
            binary,
            "tracer_symbolic_memory_dependencies_liveness",
            pov_file,
            output_initial_bytes,
        )

    def test_symbolic_cgc_transmit_handling_in_native_interface(self):
        """
        Check if CGC transmit syscall with symbolic arguments is handled in native interface when tracing.
        """

        binary = os.path.join(bin_location, "tests", "cgc", "CROMU_00008")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "CROMU_00008_POV_00000.xml")
        output_initial_bytes = (
            b"> You logged in.\n> First name: Last name: User name: Birthdate (mm/dd/yy hh:mm:ss): "
            b"Date is: 12/21/1983 5:43:21\nData added, record 0\n"
            b"> Enter search express (firstname or fn, lastname or ln, username or un, birthdate or bd,"
            b" operators ==, !=, >, <, AND and OR):\n"
        )
        add_options = {
            angr.options.UNICORN_HANDLE_CGC_RECEIVE_SYSCALL,
            angr.options.UNICORN_HANDLE_SYMBOLIC_ADDRESSES,
            angr.options.UNICORN_HANDLE_SYMBOLIC_CONDITIONS,
            angr.options.UNICORN_HANDLE_SYMBOLIC_SYSCALLS,
        }
        trace_cgc_with_pov_file(
            binary,
            "tracer_symbolic_cgc_transmit_handling_in_native_interface",
            pov_file,
            output_initial_bytes,
            add_options=add_options,
        )

    def test_user_controlled_code_execution(self):
        """
        Test user controlled code execution where instruction pointer is concrete and code is symbolic
        """
        binary = os.path.join(bin_location, "tests", "cgc", "NRFIN_00034")
        pov_file = os.path.join(bin_location, "tests_data", "cgc_povs", "NRFIN_00034_POV_00000.xml")
        output_initial_bytes = b"\x00" * 8
        trace_cgc_with_pov_file(binary, "tracer_user_controlled_code_execution", pov_file, output_initial_bytes)


if __name__ == "__main__":
    logging.getLogger("angr.simos").setLevel("DEBUG")
    logging.getLogger("angr.state_plugins.preconstrainer").setLevel("DEBUG")
    logging.getLogger("angr.exploration_techniques.tracer").setLevel("DEBUG")
    logging.getLogger("angr.exploration_techniques.crash_monitor").setLevel("DEBUG")

    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.procedures.libc"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location

test_location = os.path.join(bin_location, "tests")


class TestSprintf(unittest.TestCase):
    def test_sprintf(self):
        p = angr.Project(
            os.path.join(test_location, "x86_64", "sprintf_test"),
            auto_load_libs=False,
        )
        a = p.factory.simulation_manager().explore(find=0x4005C0)
        state = a.found[0]

        str1 = state.solver.eval(state.memory.load(0x600AD0, 13), cast_to=bytes)
        assert str1 == b"Immediate: 3\n"

        str2 = state.solver.eval(state.memory.load(0x600A70, 7), cast_to=bytes)
        assert str2 == b"Int: 3\n"

        str3 = state.solver.eval(state.memory.load(0x600AB0, 8), cast_to=bytes)
        assert str3 == b"Char: c\n"

        str4 = state.solver.eval(state.memory.load(0x600A50, 14), cast_to=bytes)
        assert str4 == b"Uninit int: 0\n"

        str5 = state.solver.eval(state.memory.load(0x600A90, 24), cast_to=bytes)
        assert str5 == b"Str on stack: A string.\n"

        str6 = state.solver.eval(state.memory.load(0x600A30, 21), cast_to=bytes)
        assert str6 == b"Global str: GLOB_STR\n"


if __name__ == "__main__":
    unittest.main()



class SimActLocation:
    """
    Structure-like class used to bundle the instruction address and statement index of a given SimAction in order to
    uniquely identify a given SimAction
    """

    def __init__(self, bbl_addr: int, ins_addr: int, stmt_idx: int):
        self.bbl_addr = bbl_addr
        self.ins_addr = ins_addr
        self.stmt_idx = stmt_idx

    def __repr__(self):
        return f"SimActLocation<{hex(self.bbl_addr)}.{hex(self.ins_addr)}.{hex(self.stmt_idx)}>"

    def __hash__(self):
        return hash((self.bbl_addr, self.ins_addr, self.stmt_idx))

    def __eq__(self, other):
        if not isinstance(other, SimActLocation):
            return False
        return self.bbl_addr == other.bbl_addr and self.ins_addr == other.ins_addr and self.stmt_idx == other.stmt_idx

    # def __add__(self, other):
    #     if not isinstance(other, int):
    #         return
    #
    #     self._stmt_idx += other


DEFAULT_LOCATION = SimActLocation(0, 0, 0)  # To be used when a location isn't necessary (eg, ConstantDepNode)


class ParsedInstruction:
    """
    Used by parser to facilitate linking with recent ancestors in an efficient manner
    """

    def __init__(
        self,
        ins_addr: int,  # Instruction that was parsed
        min_stmt_idx: int,  # Index of first statement in instruction
        max_stmt_idx: int,
    ):  # Index of last statement in instruction
        self.ins_addr = ins_addr
        self.min_stmt_idx = min_stmt_idx
        self.max_stmt_idx = max_stmt_idx



# pylint:disable=arguments-differ,unused-argument,missing-class-docstring
import angr


class access(angr.SimProcedure):
    def run(self, pathname, mode):
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]
        p_strlen = self.inline_call(strlen, pathname)
        p_expr = self.state.memory.load(pathname, p_strlen.max_null_index, endness="Iend_BE")
        path = self.state.solver.eval(p_expr, cast_to=bytes)

        simfile = self.state.fs.get(path)
        if simfile is None:
            # the file does not exist
            return -1

        return 0



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.engines"  # pylint:disable=redefined-builtin

import gc
import os
import pickle
import re
import unittest

import angr
from angr import options as so

from ..common import bin_location, broken, slow_test

test_location = os.path.join(bin_location, "tests")


def _remove_addr_from_trace_item(trace_item_str):
    m = re.match(r"(<\S+ \S+) from 0x[0-9a-f]+(:[\s\S]+)", trace_item_str)
    if m is None:
        return None
    return m.group(1) + m.group(2)


def _compare_trace(trace, expected):
    assert len(trace) == len(expected)

    for trace_item, expected_str in zip(trace, expected):
        trace_item_str = str(trace_item)
        if trace_item_str.startswith("<SimProcedure"):
            # we do not care if addresses of SimProcedures match, since they are not allocated in a deterministic way
            trace_item_str = _remove_addr_from_trace_item(trace_item_str)
            expected_str = _remove_addr_from_trace_item(expected_str)

        assert trace_item_str == expected_str


class TestUnicorn(unittest.TestCase):
    def test_stops(self):
        p = angr.Project(os.path.join(test_location, "i386", "uc_stop"), auto_load_libs=False)

        # test STOP_NORMAL, STOP_STOPPOINT
        s_normal = p.factory.entry_state(args=["a"], add_options=so.unicorn)
        s_normal.unicorn.max_steps = 100
        pg_normal = p.factory.simulation_manager(s_normal).run()
        p_normal = pg_normal.one_deadended
        _compare_trace(
            p_normal.history.descriptions,
            [
                "<Unicorn (STOP_STOPPOINT after 4 steps) from 0x8048340: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8119990: 1 sat>",
                "<Unicorn (STOP_STOPPOINT after 14 steps) from 0x8048650: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8400044: 1 sat>",
                "<Unicorn (STOP_NORMAL after 100 steps) from 0x80485b5: 1 sat>",
                "<Unicorn (STOP_STOPPOINT after 12 steps) from 0x804846f: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8400048: 1 sat>",
            ],
        )

        s_normal_angr = p.factory.entry_state(args=["a"])
        pg_normal_angr = p.factory.simulation_manager(s_normal_angr).run()
        p_normal_angr = pg_normal_angr.one_deadended
        assert p_normal_angr.history.bbl_addrs.hardcopy == p_normal.history.bbl_addrs.hardcopy

        # test STOP_STOPPOINT on an address that is not a basic block start
        s_stoppoints = p.factory.call_state(
            p.loader.find_symbol("main").rebased_addr, 1, angr.PointerWrapper([]), add_options=so.unicorn
        )

        # this address is right before/after the bb for the stop_normal() function ends
        # we should not stop there, since that code is never hit
        stop_fake = [0x0804847C, 0x08048454]

        # this is an address inside main that is not the beginning of a basic block. we should stop here
        stop_in_bb = 0x08048638
        stop_bb = 0x08048633  # basic block of the above address
        pg_stoppoints = p.factory.simulation_manager(s_stoppoints).run(n=1, extra_stop_points=stop_fake + [stop_in_bb])
        assert len(pg_stoppoints.active) == 1
        p_stoppoints = pg_stoppoints.one_active
        assert p_stoppoints.addr == stop_bb
        _compare_trace(
            p_stoppoints.history.descriptions,
            ["<Unicorn (STOP_STOPPOINT after 111 steps) from 0x80485b5: 1 sat>"],
        )

        # test STOP_SYMBOLIC_READ_SYMBOLIC_TRACKING_DISABLED
        s_symbolic_read_tracking_disabled = p.factory.entry_state(
            args=["a", "a"],
            add_options=so.unicorn,
            remove_options={so.UNICORN_SYM_REGS_SUPPORT},
        )
        pg_symbolic_read_tracking_disabled = p.factory.simulation_manager(s_symbolic_read_tracking_disabled).run()
        p_symbolic_read_tracking_disabled = pg_symbolic_read_tracking_disabled.one_deadended
        _compare_trace(
            p_symbolic_read_tracking_disabled.history.descriptions,
            [
                "<Unicorn (STOP_STOPPOINT after 4 steps) from 0x8048340: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8119990: 1 sat>",
                "<Unicorn (STOP_STOPPOINT after 14 steps) from 0x8048650: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8400044: 1 sat>",
                "<Unicorn (STOP_SYMBOLIC_READ_SYMBOLIC_TRACKING_DISABLED after 7 steps) from 0x80485b5: 1 sat>",
                "<IRSB from 0x804848a: 1 sat 3 unsat>",
                "<Unicorn (STOP_STOPPOINT after 3 steps) from 0x80484bb: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8400048: 1 sat>",
            ],
        )

        s_symbolic_read_tracking_disabled_angr = p.factory.entry_state(args=["a", "a"])
        pg_symbolic_read_tracking_disabled_angr = p.factory.simulation_manager(
            s_symbolic_read_tracking_disabled_angr
        ).run()
        p_symbolic_read_tracking_disabled_angr = pg_symbolic_read_tracking_disabled_angr.one_deadended
        assert (
            p_symbolic_read_tracking_disabled_angr.history.bbl_addrs.hardcopy
            == p_symbolic_read_tracking_disabled.history.bbl_addrs.hardcopy
        )

        # test STOP_SEGFAULT
        s_segfault = p.factory.entry_state(
            args=["a", "a", "a", "a", "a", "a", "a"],
            add_options=so.unicorn | {so.STRICT_PAGE_ACCESS, so.ENABLE_NX},
        )
        pg_segfault = p.factory.simulation_manager(s_segfault).run()
        p_segfault = pg_segfault.errored[0].state
        # TODO: fix the permissions segfault to commit if it's a MEM_FETCH
        # this will extend the last simunicorn one more block
        _compare_trace(
            p_segfault.history.descriptions,
            [
                "<Unicorn (STOP_STOPPOINT after 4 steps) from 0x8048340: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8119990: 1 sat>",
                "<Unicorn (STOP_STOPPOINT after 14 steps) from 0x8048650: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8400044: 1 sat>",
                "<Unicorn (STOP_SEGFAULT after 7 steps) from 0x80485b5: 1 sat>",
                "<IRSB from 0x8048508: 1 sat>",
            ],
        )

        s_segfault_angr = p.factory.entry_state(
            args=["a", "a", "a", "a", "a", "a", "a"],
            add_options={so.STRICT_PAGE_ACCESS, so.ENABLE_NX},
        )
        pg_segfault_angr = p.factory.simulation_manager(s_segfault_angr).run()
        p_segfault_angr = pg_segfault_angr.errored[0].state
        assert p_segfault_angr.history.bbl_addrs.hardcopy == p_segfault.history.bbl_addrs.hardcopy
        assert pg_segfault_angr.errored[0].error.addr == pg_segfault.errored[0].error.addr

        # test STOP_SYMBOLIC_BLOCK_EXIT
        s_symbolic_exit = p.factory.entry_state(args=["a"] * 10, add_options=so.unicorn)
        pg_symbolic_exit = p.factory.simulation_manager(s_symbolic_exit).run()
        p_symbolic_exit = pg_symbolic_exit.one_deadended
        _compare_trace(
            p_symbolic_exit.history.descriptions,
            [
                "<Unicorn (STOP_STOPPOINT after 4 steps) from 0x8048340: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8119990: 1 sat>",
                "<Unicorn (STOP_STOPPOINT after 14 steps) from 0x8048650: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8400044: 1 sat>",
                "<Unicorn (STOP_SYMBOLIC_BLOCK_EXIT_CONDITION after 7 steps) from 0x80485b5: 1 sat>",
                "<IRSB from 0x804855d: 2 sat 1 unsat>",
                "<Unicorn (STOP_STOPPOINT after 4 steps) from 0x8048587: 1 sat>",
                "<SimProcedure __libc_start_main from 0x8400048: 1 sat>",
            ],
        )

        s_symbolic_exit_angr = p.factory.entry_state(args=["a"] * 10)
        pg_symbolic_exit_angr = p.factory.simulation_manager(s_symbolic_exit_angr).run()
        p_symbolic_exit_angr = pg_symbolic_exit_angr.one_deadended
        assert p_symbolic_exit_angr.history.bbl_addrs.hardcopy == p_symbolic_exit.history.bbl_addrs.hardcopy

    @staticmethod
    def _run_longinit(arch):
        p = angr.Project(os.path.join(test_location, arch, "longinit"), auto_load_libs=False)
        s_unicorn = p.factory.entry_state(add_options=so.unicorn, remove_options={so.SHORT_READS})
        pg = p.factory.simulation_manager(s_unicorn, save_unconstrained=True, save_unsat=True)
        pg.explore()
        s = pg.deadended[0]
        (first, _), (second, _) = s.posix.stdin.content
        s.add_constraints(first == s.solver.BVV(b"A" * 9))
        s.add_constraints(second == s.solver.BVV(b"B" * 9))
        assert s.posix.dumps(1) == b"You entered AAAAAAAAA and BBBBBBBBB!\n"

    def test_longinit_i386(self):
        self._run_longinit("i386")

    def test_longinit_x86_64(self):
        self._run_longinit("x86_64")

    @broken
    def test_fauxware_arm(self):
        p = angr.Project(os.path.join(test_location, "armel", "fauxware"), auto_load_libs=False)
        s_unicorn = p.factory.entry_state(add_options=so.unicorn)  # unicorn
        pg = p.factory.simulation_manager(s_unicorn)
        pg.explore()
        assert all("Unicorn" in "".join(p.history.descriptions.hardcopy) for p in pg.deadended)
        assert sorted(pg.mp_deadended.posix.dumps(1).mp_items) == sorted(
            (
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
                b"Username: \nPassword: \nGo away!",
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
            )
        )

    def test_fauxware(self):
        p = angr.Project(os.path.join(test_location, "i386", "fauxware"), auto_load_libs=False)
        s_unicorn = p.factory.entry_state(add_options=so.unicorn)  # unicorn
        pg = p.factory.simulation_manager(s_unicorn)
        pg.explore()

        assert all("Unicorn" in "".join(p.history.descriptions.hardcopy) for p in pg.deadended)
        assert sorted(pg.mp_deadended.posix.dumps(1).mp_items) == sorted(
            (
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
                b"Username: \nPassword: \nGo away!",
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
            )
        )

    def test_fauxware_aggressive(self):
        p = angr.Project(os.path.join(test_location, "i386", "fauxware"), auto_load_libs=False)
        s_unicorn = p.factory.entry_state(
            add_options=so.unicorn | {so.UNICORN_AGGRESSIVE_CONCRETIZATION},
            remove_options={so.LAZY_SOLVES},
        )  # unicorn
        s_unicorn.unicorn.cooldown_symbolic_stop = 2
        s_unicorn.unicorn.cooldown_unsupported_stop = 2
        s_unicorn.unicorn.cooldown_nonunicorn_blocks = 0

        pg = p.factory.simulation_manager(s_unicorn)
        pg.explore()

        assert len(pg.deadended) == 1

    def test_partial_reads(self):
        """
        This test case if unicorn engine correctly handles case when symbolic taint is introduced by the second partial
        read performed by unicorn. Unicorn triggers memory read hook twice when reading value greater than 8 bytes on
        x86-64.
        """

        p = angr.Project(
            os.path.join(test_location, "x86_64", "test_partial_reads_handling_in_unicorn"),
            auto_load_libs=False,
        )
        # Do not treat as uninitialized memory as symbolic. Prevents introducing undesired symbolic taint
        init_state = p.factory.full_init_state(add_options=so.unicorn | {so.ZERO_FILL_UNCONSTRAINED_MEMORY})
        global_var_val = [
            init_state.solver.BVV(0x41414141, 32),
            init_state.solver.BVV(0x42424242, 32),
            init_state.solver.BVS("symb_val_0", 32),
            init_state.solver.BVS("symb_val_1", 32),
        ]
        global_var_symb = p.loader.find_symbol("global_var")
        # Store every byte separately so that entire variable is not treated as symbolic
        for count, val in enumerate(global_var_val):
            init_state.memory.store(
                global_var_symb.rebased_addr + count * 4, val, endness=init_state.arch.memory_endness
            )

        pg = p.factory.simulation_manager(init_state)
        pg.run()
        assert len(pg.deadended) == 1

    @staticmethod
    def _run_similarity(binpath, depth, prehook=None):
        b = angr.Project(os.path.join(test_location, binpath), auto_load_libs=False)
        cc = b.analyses.CongruencyCheck(throw=True)
        cc.set_state_options(
            left_add_options=so.unicorn,
            left_remove_options={
                so.LAZY_SOLVES,
                so.TRACK_MEMORY_MAPPING,
                so.COMPOSITE_SOLVER,
            },
            right_add_options={so.ZERO_FILL_UNCONSTRAINED_REGISTERS},
            right_remove_options={
                so.LAZY_SOLVES,
                so.TRACK_MEMORY_MAPPING,
                so.COMPOSITE_SOLVER,
            },
        )
        if prehook:
            cc.simgr = prehook(cc.simgr)
        cc.run(depth=depth)

    @slow_test
    def test_similarity_fauxware(self):
        def cooldown(pg):
            # gotta skip the initializers because of cpuid and RDTSC
            pg.one_left.unicorn.countdown_nonunicorn_blocks = 39
            return pg

        self._run_similarity(os.path.join("i386", "fauxware"), 1000, prehook=cooldown)

    def test_fp(self):
        with open(os.path.join(bin_location, "tests_src", "manyfloatsum.c"), encoding="utf-8") as fp:
            type_cache = angr.sim_type.parse_defns(fp.read())
        p = angr.Project(os.path.join(test_location, "i386", "manyfloatsum"), auto_load_libs=False)

        for function in (
            "sum_floats",
            "sum_combo",
            "sum_segregated",
            "sum_doubles",
            "sum_combo_doubles",
            "sum_segregated_doubles",
        ):
            args = list(range(len(type_cache[function].args)))
            answer = float(sum(args))
            addr = p.loader.find_symbol(function).rebased_addr
            my_callable = p.factory.callable(addr, prototype=type_cache[function])
            my_callable.set_base_state(p.factory.blank_state(add_options=so.unicorn))
            result = my_callable(*args)
            assert not result.symbolic
            result_concrete = result.args[0]
            assert answer == result_concrete

    def test_unicorn_pickle(self):
        p = angr.Project(os.path.join(test_location, "i386", "fauxware"), auto_load_libs=False)

        def _uni_state():
            # try pickling out paths that went through unicorn
            s_unicorn = p.factory.entry_state(add_options=so.unicorn)
            s_unicorn.unicorn.countdown_nonunicorn_blocks = 0
            s_unicorn.unicorn.countdown_symbolic_stop = 0
            s_unicorn.unicorn.cooldown_nonunicorn_blocks = 0
            s_unicorn.unicorn.cooldown_symbolic_stop = 2
            return s_unicorn

        pg = p.factory.simulation_manager(_uni_state())
        pg.one_active.options.update(so.unicorn)
        pg.run(until=lambda lpg: "Unicorn" in lpg.one_active.history.recent_description)
        assert len(pg.active) > 0

        pgp = pickle.dumps(pg, -1)
        del pg
        gc.collect()
        pg2 = pickle.loads(pgp)
        pg2.explore()

        assert sorted(pg2.mp_deadended.posix.dumps(1).mp_items) == sorted(
            (
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
                b"Username: \nPassword: \nGo away!",
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
            )
        )

        # test the pickling of SimUnicorn itself
        p = angr.Project(os.path.join(test_location, "i386", "fauxware"), auto_load_libs=False)
        pg = p.factory.simulation_manager(_uni_state())
        pg.run(n=2)
        assert p.factory.successors(pg.one_active).sort == "Unicorn"

        pgp = pickle.dumps(pg, -1)
        del pg
        gc.collect()
        pg2 = pickle.loads(pgp)
        pg2.explore()

        assert sorted(pg2.mp_deadended.posix.dumps(1).mp_items) == sorted(
            (
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
                b"Username: \nPassword: \nGo away!",
                b"Username: \nPassword: \nWelcome to the admin console, trusted user!\n",
            )
        )

    def test_concrete_transmits(self):
        p = angr.Project(os.path.join(test_location, "cgc", "PIZZA_00001"), auto_load_libs=False)
        inp = bytes.fromhex("320a310a0100000005000000330a330a340a")

        s_unicorn = p.factory.entry_state(
            add_options=so.unicorn | {so.CGC_NO_SYMBOLIC_RECEIVE_LENGTH},
            stdin=inp,
            flag_page=b"\0" * 4096,
        )
        pg_unicorn = p.factory.simulation_manager(s_unicorn)
        pg_unicorn.run(n=10)

        assert pg_unicorn.one_active.posix.dumps(1) == (
            b"1) Add number to the array\n"
            b"2) Add random number to the array\n"
            b"3) Sum numbers\n"
            b"4) Exit\n"
            b"Randomness added\n"
            b"1) Add number to the array\n"
            b"2) Add random number to the array\n"
            b"3) Sum numbers\n"
            b"4) Exit\n"
            b"  Index: \n"
            b"1) Add number to the array\n"
            b"2) Add random number to the array\n"
            b"3) Sum numbers\n"
            b"4) Exit\n"
        )

    def test_inspect(self):
        p = angr.Project(os.path.join(test_location, "i386", "uc_stop"), auto_load_libs=False)

        def main_state(argc, add_options=None):
            add_options = add_options or so.unicorn
            main_addr = p.loader.find_symbol("main").rebased_addr
            return p.factory.call_state(main_addr, argc, angr.PointerWrapper([]), add_options=add_options)

        # test breaking on specific addresses
        s_break_addr = main_state(1)
        addr0 = 0x08048479  # at the beginning of a basic block, at end of stop_normal function
        addr1 = 0x080485D0  # this is at the beginning of main, in the middle of a basic block
        addr2 = 0x08048461  # another non-bb address, at the start of stop_normal
        addr3 = 0x0804847C  # address of a block that should not get hit (stop_symbolc function)
        addr4 = 0x08048632  # another address that shouldn't get hit, near end of main
        hits = {addr0: 0, addr1: 0, addr2: 0, addr3: 0, addr4: 0}

        def create_addr_action(addr):
            def action(_state):
                hits[addr] += 1

            return action

        for addr in [addr0, addr1, addr2]:
            s_break_addr.inspect.b("instruction", instruction=addr, action=create_addr_action(addr))

        pg_instruction = p.factory.simulation_manager(s_break_addr)
        pg_instruction.run()
        assert hits[addr0] == 1
        assert hits[addr1] == 1
        assert hits[addr2] == 1
        assert hits[addr3] == 0
        assert hits[addr4] == 0

        # test breaking on every instruction
        def collect_trace(options):
            s_break_every = main_state(1, add_options=options)
            trace = []

            def action_every(state):
                trace.append(state.addr)

            s_break_every.inspect.b("instruction", action=action_every)
            pg_break_every = p.factory.simulation_manager(s_break_every)
            pg_break_every.run()

        assert collect_trace(so.unicorn) == collect_trace(set())

    def test_explore(self):
        p = angr.Project(os.path.join(test_location, "i386", "uc_stop"), auto_load_libs=False)

        def main_state(argc, add_options=None):
            add_options = add_options or so.unicorn
            main_addr = p.loader.find_symbol("main").rebased_addr
            return p.factory.call_state(main_addr, argc, angr.PointerWrapper([]), add_options=add_options)

        addr = 0x08048479
        s_explore = main_state(1)
        pg_explore_find = p.factory.simulation_manager(s_explore)
        pg_explore_find.explore(find=addr)
        assert len(pg_explore_find.found) == 1
        assert pg_explore_find.found[0].addr == addr

        pg_explore_avoid = p.factory.simulation_manager(s_explore)
        pg_explore_avoid.explore(avoid=addr)
        assert len(pg_explore_avoid.avoid) == 1
        assert pg_explore_avoid.avoid[0].addr == addr

    def test_single_step(self):
        p = angr.Project(os.path.join(test_location, "i386", "uc_stop"), auto_load_libs=False)

        def main_state(argc, add_options=None):
            add_options = add_options or so.unicorn
            main_addr = p.loader.find_symbol("main").rebased_addr
            return p.factory.call_state(main_addr, argc, angr.PointerWrapper([]), add_options=add_options)

        s_main = main_state(1)

        step1 = s_main.block().instruction_addrs[1]
        successors1 = s_main.step(num_inst=1).successors
        assert len(successors1) == 1
        assert successors1[0].addr == step1

        step5 = s_main.block().instruction_addrs[5]
        successors2 = successors1[0].step(num_inst=4).successors
        assert len(successors2) == 1
        assert successors2[0].addr == step5

    def test_symbolic_flags_preserved_on_stop(self):
        """
        Test if symbolic flags are preserved when unicorn engine stops. This is needed for cases where compare is
        performed in one block and conditional jump in another.
        """

        p = angr.Project(os.path.join(test_location, "x86_64", "test_symbolic_flags_in_unicorn"))
        init_state = p.factory.full_init_state(add_options=angr.options.unicorn)
        simgr = p.factory.simgr(init_state)
        simgr.run()
        result = None
        for final_state in simgr.deadended:
            if b"Congrats" in final_state.posix.dumps(1):
                result = final_state.posix.dumps(0)
                break

        assert result == b"FLAG{l00ps_4r3_t00_34sy_r1gh7??}"


if __name__ == "__main__":
    import logging

    logging.getLogger("angr.state_plugins.unicorn_engine").setLevel("DEBUG")
    logging.getLogger("angr.engines.unicorn_engine").setLevel("INFO")
    logging.getLogger("angr.factory").setLevel("DEBUG")
    logging.getLogger("angr.project").setLevel("DEBUG")

    unittest.main()



import angr
import logging

l = logging.getLogger("angr.procedures.win32.gethostbyname")


class gethostbyname(angr.SimProcedure):
    def run(self, _):  # pylint:disable=arguments-differ
        ret_expr = self.state.solver.BVS("gethostbyname_retval", 32, key=("api", "gethostbyname_retval"))
        return ret_expr



import angr


class ReturnChar(angr.SimProcedure):
    def run(self):
        s_var = self.state.solver.Unconstrained("char_ret", self.state.arch.bits, key=("api", "?", self.display_name))
        self.state.add_constraints(
            self.state.solver.And(self.state.solver.ULE(s_var, 126), self.state.solver.UGE(s_var, 9))
        )
        return s_var



#!/usr/bin/env python3
__package__ = __package__ or "tests.analyses.reaching_definitions"  # pylint:disable=redefined-builtin

import os
import random

from unittest import main, mock, TestCase

import archinfo

from angr.analyses.reaching_definitions.heap_allocator import HeapAllocator
from angr.analyses.reaching_definitions.rd_state import ReachingDefinitionsState
from angr.analyses.reaching_definitions.subject import SubjectType
from angr.knowledge_plugins.key_definitions.live_definitions import LiveDefinitions
from angr.code_location import CodeLocation

from ...common import bin_location


TESTS_LOCATION = os.path.join(bin_location, "tests")


class _MockFunctionSubject:  # pylint:disable=too-few-public-methods
    class _MockFunction:  # pylint:disable=too-few-public-methods
        def __init__(self):
            self.addr = 0x42

    def __init__(self):
        self.type = SubjectType.Function
        self.cc = None  # pylint:disable=invalid-name
        self.content = self._MockFunction()


class TestReachingDefinitionsState(TestCase):
    def test_initializing_rd_state_for_ppc_without_rtoc_value_should_raise_an_error(self):
        arch = archinfo.arch_ppc64.ArchPPC64()
        self.assertRaises(
            ValueError, ReachingDefinitionsState, CodeLocation(0x42, None), arch=arch, subject=_MockFunctionSubject()
        )

    def test_initializing_rd_state_for_ppc_with_rtoc_value(self):
        arch = archinfo.arch_ppc64.ArchPPC64()
        rtoc_value = random.randint(0, 0xFFFFFFFFFFFFFFFF)

        state = ReachingDefinitionsState(
            CodeLocation(0x42, None), arch=arch, subject=_MockFunctionSubject(), rtoc_value=rtoc_value
        )

        rtoc_offset = arch.registers["rtoc"][0]
        rtoc_definition_value = state.registers.load(rtoc_offset, size=8)

        self.assertIsNotNone(rtoc_definition_value.one_value())
        v = rtoc_definition_value.one_value()
        self.assertFalse(v.symbolic)
        self.assertEqual(v.concrete_value, rtoc_value)

    def test_rd_state_gets_a_default_heap_allocator(self):
        arch = archinfo.arch_arm.ArchARM()
        state = ReachingDefinitionsState(CodeLocation(0x42, None), arch, _MockFunctionSubject())

        self.assertTrue(isinstance(state.heap_allocator, HeapAllocator))

    def test_get_sp_delegates_to_the_underlying_live_definitions(self):  # pylint:disable=no-self-use
        arch = archinfo.arch_arm.ArchARM()
        live_definitions = LiveDefinitions(arch)

        state = ReachingDefinitionsState(
            CodeLocation(0x42, None), arch=arch, subject=_MockFunctionSubject(), live_definitions=live_definitions
        )

        with mock.patch.object(LiveDefinitions, "get_sp") as live_definitions_get_sp_mock:
            state.get_sp()

            live_definitions_get_sp_mock.assert_called_once()


if __name__ == "__main__":
    main()



# pylint:disable=wrong-import-position,wrong-import-order
from typing import Optional, List, Tuple, Union
import logging
from collections import defaultdict

import networkx

import claripy
import pyvex
import ailment

import angr.errors
from ...storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
from ...block import Block
from ...errors import AngrVariableRecoveryError, SimEngineError
from ...knowledge_plugins import Function
from ...sim_variable import SimStackVariable, SimRegisterVariable, SimVariable, SimMemoryVariable
from ...engines.vex.claripy.irop import vexop_to_simop
from angr.analyses import ForwardAnalysis, visitors
from ..typehoon.typevars import Equivalence, TypeVariable
from .variable_recovery_base import VariableRecoveryBase, VariableRecoveryStateBase
from .engine_vex import SimEngineVRVEX
from .engine_ail import SimEngineVRAIL

l = logging.getLogger(name=__name__)


class VariableRecoveryFastState(VariableRecoveryStateBase):
    """
    The abstract state of variable recovery analysis.

    :ivar KeyedRegion stack_region: The stack store.
    :ivar KeyedRegion register_region:  The register store.
    """

    def __init__(
        self,
        block_addr,
        analysis,
        arch,
        func,
        stack_region=None,
        register_region=None,
        global_region=None,
        typevars=None,
        type_constraints=None,
        delayed_type_constraints=None,
        stack_offset_typevars=None,
        project=None,
        ret_val_size=None,
    ):
        super().__init__(
            block_addr,
            analysis,
            arch,
            func,
            stack_region=stack_region,
            register_region=register_region,
            global_region=global_region,
            typevars=typevars,
            type_constraints=type_constraints,
            delayed_type_constraints=delayed_type_constraints,
            stack_offset_typevars=stack_offset_typevars,
            project=project,
        )
        self.ret_val_size = ret_val_size

    def __repr__(self):
        return "<VRAbstractState@%#x: %d register variables, %d stack variables>" % (
            self.block_addr,
            len(self.register_region),
            len(self.stack_region),
        )

    def __eq__(self, other):
        if type(other) is not VariableRecoveryFastState:
            return False
        return self.stack_region == other.stack_region and self.register_region == other.register_region

    def copy(self):
        state = VariableRecoveryFastState(
            self.block_addr,
            self._analysis,
            self.arch,
            self.function,
            stack_region=self.stack_region.copy(),
            register_region=self.register_region.copy(),
            global_region=self.global_region.copy(),
            typevars=self.typevars.copy(),
            type_constraints=self.type_constraints.copy(),
            delayed_type_constraints=self.delayed_type_constraints.copy(),
            stack_offset_typevars=dict(self.stack_offset_typevars),
            project=self.project,
            ret_val_size=self.ret_val_size,
        )

        return state

    def merge(
        self, others: Tuple["VariableRecoveryFastState"], successor=None
    ) -> Tuple["VariableRecoveryFastState", bool]:
        """
        Merge two abstract states.

        For any node A whose dominance frontier that the current node (at the current program location) belongs to, we
        create a phi variable V' for each variable V that is defined in A, and then replace all existence of V with V'
        in the merged abstract state.

        :param others: Other abstract states to merge.
        :return:       The merged abstract state.
        """

        self.phi_variables = {}  # A mapping from original variable and its corresponding phi variable
        self.successor_block_addr = successor

        merged_stack_region = self.stack_region.copy()
        merged_stack_region.set_state(self)
        merge_occurred = merged_stack_region.merge([other.stack_region for other in others], None)

        merged_register_region = self.register_region.copy()
        merged_register_region.set_state(self)
        merge_occurred |= merged_register_region.merge([other.register_region for other in others], None)

        merged_global_region = self.global_region.copy()
        merged_global_region.set_state(self)
        merge_occurred |= merged_global_region.merge([other.global_region for other in others], None)

        merged_typevars = self.typevars
        merged_typeconstraints = self.type_constraints.copy()
        delayed_typeconstraints = self.delayed_type_constraints.copy().clean()
        for other in others:
            merged_typevars = merged_typevars.merge(other.typevars)
            merged_typeconstraints |= other.type_constraints
            for v, cons in other.delayed_type_constraints.items():
                delayed_typeconstraints[v] |= cons

        merge_occurred |= self.typevars != merged_typevars
        merge_occurred |= self.type_constraints != merged_typeconstraints
        merge_occurred |= self.delayed_type_constraints != delayed_typeconstraints

        # add subtype constraints for all replacements
        for v0, v1 in self.phi_variables.items():
            # v0 will be replaced by v1
            if not merged_typevars.has_type_variable_for(v1, None):
                merged_typevars.add_type_variable(v1, None, TypeVariable())
            if not merged_typevars.has_type_variable_for(v0, None):
                merged_typevars.add_type_variable(v0, None, TypeVariable())
            # Assuming v2 = phi(v0, v1), then we know that v0_typevar == v1_typevar == v2_typevar
            # However, it's possible that neither v0 nor v1 will ever be used in future blocks, which not only makes
            # this phi function useless, but also leads to the incorrect assumption that v1_typevar == v2_typevar.
            # Hence, we delay the addition of the equivalence relationship into the type constraints. It is only added
            # when v1 (the new variable that will end up in the state) is ever used in the future.

            # create an equivalence relationship
            equivalence = Equivalence(
                merged_typevars.get_type_variable(v1, None), merged_typevars.get_type_variable(v0, None)
            )
            delayed_typeconstraints[v1].add(equivalence)

        stack_offset_typevars = {}
        all_stack_addr_typevar_offsets = set(self.stack_offset_typevars)
        for other in others:
            all_stack_addr_typevar_offsets.update(other.stack_offset_typevars)
        for offset in all_stack_addr_typevar_offsets:
            all_typevars = set()
            for state in (self,) + others:
                typevar = state.stack_offset_typevars.get(offset, None)
                if typevar is not None:
                    all_typevars.add(typevar)

            if len(all_typevars) == 1:
                typevar = all_typevars.pop()
            else:
                typevar = TypeVariable()
                for orig_typevar in all_typevars:
                    merged_typeconstraints.add(Equivalence(orig_typevar, typevar))
            stack_offset_typevars[offset] = typevar

        ret_val_size = self.ret_val_size
        for o in others:
            if o.ret_val_size is not None:
                if ret_val_size is None or o.ret_val_size > ret_val_size:
                    ret_val_size = o.ret_val_size
                    merge_occurred = True

        # clean up
        self.phi_variables = {}
        self.successor_block_addr = None

        state = VariableRecoveryFastState(
            successor,
            self._analysis,
            self.arch,
            self.function,
            stack_region=merged_stack_region,
            register_region=merged_register_region,
            global_region=merged_global_region,
            typevars=merged_typevars,
            type_constraints=merged_typeconstraints,
            delayed_type_constraints=delayed_typeconstraints,
            stack_offset_typevars=stack_offset_typevars,
            project=self.project,
            ret_val_size=ret_val_size,
        )

        return state, merge_occurred

    #
    # Util methods
    #

    def _normalize_register_offset(self, offset):  # pylint:disable=no-self-use
        # TODO:

        return offset

    def _to_signed(self, n):
        if n >= 2 ** (self.arch.bits - 1):
            # convert it to a negative number
            return n - 2**self.arch.bits

        return n


class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:disable=abstract-method
    """
    Recover "variables" from a function by keeping track of stack pointer offsets and pattern matching VEX statements.

    If calling conventions are recovered prior to running VariableRecoveryFast, variables can be recognized more
    accurately. However, it is not a requirement. In this case, the function graph you pass must contain information
    indicating the call-out sites inside the analyzed function. These graph edges must be annotated with either
    ``"type": "call"`` or ``"outside": True``.
    """

    def __init__(
        self,
        func: Union[Function, str, int],
        func_graph: Optional[networkx.DiGraph] = None,
        max_iterations: int = 2,
        low_priority=False,
        track_sp=True,
        func_args: Optional[List[SimVariable]] = None,
        store_live_variables=False,
        unify_variables=True,
    ):
        if not isinstance(func, Function):
            func = self.kb.functions[func]
        func_graph_with_calls = func_graph or func.transition_graph
        call_info = defaultdict(list)
        for node_from, node_to, data in func_graph_with_calls.edges(data=True):
            if data.get("type", None) == "call" or data.get("outside", False):
                try:
                    call_info[node_from.addr].append(self.kb.functions.get_by_addr(node_to.addr))
                except KeyError:
                    pass

        function_graph_visitor = visitors.FunctionGraphVisitor(func, graph=func_graph)

        # Make sure the function is not empty
        if not func.block_addrs_set or func.startpoint is None:
            raise AngrVariableRecoveryError("Function %s is empty." % repr(func))

        VariableRecoveryBase.__init__(self, func, max_iterations, store_live_variables)
        ForwardAnalysis.__init__(
            self, order_jobs=True, allow_merging=True, allow_widening=False, graph_visitor=function_graph_visitor
        )

        self._low_priority = low_priority
        self._job_ctr = 0
        self._track_sp = track_sp and self.project.arch.sp_offset is not None
        self._func_args = func_args
        self._unify_variables = unify_variables

        self._ail_engine = SimEngineVRAIL(self.project, self.kb, call_info=call_info)
        self._vex_engine = SimEngineVRVEX(self.project, self.kb, call_info=call_info)

        self._node_iterations = defaultdict(int)

        self._node_to_cc = {}
        self.var_to_typevars = defaultdict(set)
        self.type_constraints = None
        self.ret_val_size = None

        self._analyze()

        # cleanup (for cpython pickle)
        self.downsize()
        self._ail_engine = None
        self._vex_engine = None

    #
    # Main analysis routines
    #

    def _pre_analysis(self):
        self.type_constraints = set()

        self.initialize_dominance_frontiers()

        if self._track_sp:
            # initialize node_to_cc map
            function_nodes = [n for n in self.function.transition_graph.nodes() if isinstance(n, Function)]
            # all nodes that end with a call must be in the _node_to_cc dict
            for func_node in function_nodes:
                for callsite_node in self.function.transition_graph.predecessors(func_node):
                    if func_node.calling_convention is None:
                        # l.warning("Unknown calling convention for %r.", func_node)
                        self._node_to_cc[callsite_node.addr] = None
                    else:
                        self._node_to_cc[callsite_node.addr] = func_node.calling_convention

    def _pre_job_handling(self, job):
        self._job_ctr += 1
        if self._low_priority:
            self._release_gil(self._job_ctr, 100, 0.000001)

    def _initial_abstract_state(self, node):
        state = VariableRecoveryFastState(
            node.addr,
            self,
            self.project.arch,
            self.function,
            project=self.project,
        )
        initial_sp = state.stack_address(self.project.arch.bytes if self.project.arch.call_pushes_ret else 0)
        if self.project.arch.sp_offset is not None:
            state.register_region.store(self.project.arch.sp_offset, initial_sp)
        # give it enough stack space
        if self.project.arch.bp_offset is not None:
            state.register_region.store(self.project.arch.bp_offset, initial_sp + 0x100000)

        internal_manager = self.variable_manager[self.function.addr]

        # put a return address on the stack if necessary
        if self.project.arch.call_pushes_ret:
            ret_addr_offset = self.project.arch.bytes
            # find existing variable
            ret_addr_var = next(
                iter(
                    v
                    for v in internal_manager.find_variables_by_stack_offset(ret_addr_offset)
                    if v.category == "return_address"
                ),
                None,
            )
            if ret_addr_var is None:
                ret_addr_var = SimStackVariable(
                    ret_addr_offset,
                    self.project.arch.bytes,
                    base="bp",
                    name="ret_addr",
                    region=self.function.addr,
                    category="return_address",
                    ident=internal_manager.next_variable_ident("stack"),
                )
            ret_addr = claripy.BVS("ret_addr", self.project.arch.bits)
            ret_addr = state.annotate_with_variables(ret_addr, [(0, ret_addr_var)])
            state.stack_region.store(
                state.stack_addr_from_offset(ret_addr_offset), ret_addr, endness=self.project.arch.memory_endness
            )
            internal_manager.add_variable("stack", ret_addr_offset, ret_addr_var)

        if self.project.arch.name.startswith("MIPS"):
            t9_offset, t9_size = self.project.arch.registers["t9"]
            try:
                t9_val = state.register_region.load(t9_offset, t9_size)
                if state.is_top(t9_val):
                    state.register_region.store(t9_offset, claripy.BVV(node.addr, t9_size * 8))
            except angr.errors.SimMemoryMissingError:
                state.register_region.store(t9_offset, claripy.BVV(node.addr, t9_size * 8))

        if self._func_args:
            for arg in self._func_args:
                if isinstance(arg, SimRegisterVariable):
                    v = claripy.BVS("reg_arg", arg.bits)
                    v = state.annotate_with_variables(v, [(0, arg)])
                    state.register_region.store(arg.reg, v)
                    internal_manager.add_variable("register", arg.reg, arg)
                elif isinstance(arg, SimStackVariable):
                    v = claripy.BVS("stack_arg", arg.bits)
                    v = state.annotate_with_variables(v, [(0, arg)])
                    state.stack_region.store(
                        state.stack_addr_from_offset(arg.offset),
                        v,
                        endness=self.project.arch.memory_endness,
                    )
                    internal_manager.add_variable("stack", arg.offset, arg)
                else:
                    raise TypeError("Unsupported function argument type %s." % type(arg))

        return state

    def _merge_states(self, node, *states: VariableRecoveryFastState):
        merged_state, merge_occurred = states[0].merge(states[1:], successor=node.addr)
        return merged_state, not merge_occurred

    def _run_on_node(self, node, state):
        """


        :param angr.Block node:
        :param VariableRecoveryState state:
        :return:
        """

        if type(node) is ailment.Block:
            # AIL mode
            block = node
            block_key = node.addr, node.idx
        else:
            # VEX mode, get the block again
            block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=False)
            block_key = node.addr

        # if node.addr in self._instates:
        #     prev_state: VariableRecoveryFastState = self._instates[node.addr]
        #     if input_state == prev_state:
        #         l.debug('Skip node %#x as we have reached a fixed-point', node.addr)
        #         return False, input_state
        #     else:
        #         l.debug('Merging input state of node %#x with the previous state.', node.addr)
        #         input_state, _ = prev_state.merge((input_state,), successor=node.addr)

        state = state.copy()
        state.block_addr = node.addr
        if isinstance(node, ailment.Block):
            state.block_idx = node.idx
        # self._instates[node.addr] = state

        if self._node_iterations[block_key] >= self._max_iterations:
            l.debug("Skip node %#x as we have iterated %d times on it.", node.addr, self._node_iterations[node.addr])
            return False, state

        self._process_block(state, block)

        self._node_iterations[block_key] += 1
        self.type_constraints |= state.type_constraints
        for var, typevar in state.typevars._typevars.items():
            self.var_to_typevars[var].add(typevar)

        if state.ret_val_size is not None:
            if self.ret_val_size is None or self.ret_val_size < state.ret_val_size:
                self.ret_val_size = state.ret_val_size

        state.downsize()
        self._outstates[block_key] = state

        return True, state

    def _intra_analysis(self):
        pass

    def _post_analysis(self):
        VariableRecoveryBase._post_analysis(self)

        self.variable_manager["global"].assign_variable_names(labels=self.kb.labels)
        self.variable_manager[self.function.addr].assign_variable_names()

        if self._store_live_variables:
            for addr, state in self._outstates.items():
                self.variable_manager[self.function.addr].set_live_variables(
                    addr,
                    state.downsize_region(state.register_region),
                    state.downsize_region(state.stack_region),
                )

        if self._unify_variables:
            self.variable_manager[self.function.addr].unify_variables()

        # unify type variables for global variables
        for var, typevars in self.var_to_typevars.items():
            if len(typevars) > 1 and isinstance(var, SimMemoryVariable) and not isinstance(var, SimStackVariable):
                sorted_typevars = list(sorted(typevars, key=lambda x: str(x)))  # pylint:disable=unnecessary-lambda
                for tv in sorted_typevars[1:]:
                    self.type_constraints.add(Equivalence(sorted_typevars[0], tv))

        self.variable_manager[self.function.addr].ret_val_size = self.ret_val_size

    #
    # Private methods
    #

    @staticmethod
    def _get_irconst(value, size):
        mapping = {
            1: pyvex.const.U1,
            8: pyvex.const.U8,
            16: pyvex.const.U16,
            32: pyvex.const.U32,
            64: pyvex.const.U64,
            128: pyvex.const.V128,
            256: pyvex.const.V256,
        }
        if size not in mapping:
            raise TypeError("Unsupported size %d." % size)
        return mapping.get(size)(value)

    def _peephole_optimize(self, block: Block):
        # find regN = xor(regN, regN) and replace it with PUT(regN) = 0
        i = 0
        while i < len(block.vex.statements) - 3:
            stmt0 = block.vex.statements[i]
            next_i = i + 1
            if isinstance(stmt0, pyvex.IRStmt.WrTmp) and isinstance(stmt0.data, pyvex.IRStmt.Get):
                stmt1 = block.vex.statements[i + 1]
                if isinstance(stmt1, pyvex.IRStmt.WrTmp) and isinstance(stmt1.data, pyvex.IRStmt.Get):
                    next_i = i + 2
                    if stmt0.data.offset == stmt1.data.offset and stmt0.data.ty == stmt1.data.ty:
                        next_i = i + 3
                        reg_offset = stmt0.data.offset
                        tmp0 = stmt0.tmp
                        tmp1 = stmt1.tmp
                        stmt2 = block.vex.statements[i + 2]
                        if isinstance(stmt2, pyvex.IRStmt.WrTmp) and isinstance(stmt2.data, pyvex.IRExpr.Binop):
                            if (
                                isinstance(stmt2.data.args[0], pyvex.IRExpr.RdTmp)
                                and isinstance(stmt2.data.args[1], pyvex.IRExpr.RdTmp)
                                and {stmt2.data.args[0].tmp, stmt2.data.args[1].tmp} == {tmp0, tmp1}
                                and vexop_to_simop(stmt2.data.op)._generic_name == "Xor"
                            ):
                                # found it!
                                # make a copy so we don't trash the cached VEX IRSB
                                block._vex = block.vex.copy()
                                block.vex.statements[i] = pyvex.IRStmt.NoOp()
                                block.vex.statements[i + 1] = pyvex.IRStmt.NoOp()
                                zero = pyvex.IRExpr.Const(self._get_irconst(0, block.vex.tyenv.sizeof(tmp0)))
                                block.vex.statements[i + 2] = pyvex.IRStmt.Put(zero, reg_offset)
            i = next_i
        return block

    def _process_block(self, state, block):  # pylint:disable=no-self-use
        """
        Scan through all statements and perform the following tasks:
        - Find stack pointers and the VEX temporary variable storing stack pointers
        - Selectively calculate VEX statements
        - Track memory loading and mark stack and global variables accordingly

        :param angr.Block block:
        :return:
        """

        l.debug("Processing block %#x.", block.addr)

        if isinstance(block, Block):
            try:
                _ = block.vex
            except SimEngineError:
                # the block does not exist or lifting failed
                return
            block = self._peephole_optimize(block)

        processor = self._ail_engine if isinstance(block, ailment.Block) else self._vex_engine
        processor.process(state, block=block, fail_fast=self._fail_fast)

        if self._track_sp and block.addr in self._node_to_cc:
            # readjusting sp at the end for blocks that end in a call
            sp: MultiValues = state.register_region.load(self.project.arch.sp_offset, size=self.project.arch.bytes)
            sp_v = sp.one_value()
            if sp_v is None:
                l.warning("Unexpected stack pointer value at the end of the function. Pick the first one.")
                sp_v = next(iter(next(iter(sp.values()))))

            adjusted = False

            if not adjusted:
                # make a guess
                # of course, this will fail miserably if the function called is not cdecl
                if self.project.arch.call_pushes_ret:
                    sp_v += self.project.arch.bytes
                    adjusted = True

            if adjusted:
                state.register_region.store(self.project.arch.sp_offset, sp_v)


from angr.analyses import AnalysesHub

AnalysesHub.register_default("VariableRecoveryFast", VariableRecoveryFast)
